import json as JSON
from confluent_kafka import Producer
from ..base import Node


class Kafka(Node):
    '''Connect to kafka server and send data

    Args:
        foo (callable): input stream
        foo_kwargs (dict): kwargs for the input stream
        servers (list): kafka bootstrap servers
        group (str): kafka group id
        topics (list): list of kafka topics to connect to
        json (bool): load input data as json
        wrap (bool): wrap result in a list
        interval (int): kafka poll interval
    '''
    def __init__(self, node, servers='', topic='', json=False, wrap=False):
        p = Producer({'bootstrap.servers': servers})

        async def _send(data, producer=p, topic=topic, json=json, wrap=wrap):
            ret = []
            # Trigger any available delivery report callbacks from previous produce() calls
            producer.poll(0)

            if wrap:
                data = [data]

            if json:
                data = JSON.dumps(data)

            producer.produce(topic, data.encode('utf-8'), callback=lambda *args: ret.append(args))

            for data in ret:
                yield data
            ret = []

        super().__init__(foo=_send, name='Kafka', inputs=1)
        node._downstream.append((self, 0))
        self._upstream.append(node)



import requests
import json as JSON
from ..base import Node
from ...base import StreamEnd


class HTTP(Node):
    '''Connect to url and post results to it

    Args:
        foo (callable): input stream
        foo_kwargs (dict): kwargs for the input stream
        url (str): url to post to
        json (bool): dump data as json
        wrap (bool): wrap input in a list
        field (str): field to index result by
        proxies (list): list of URL proxies to pass to requests.post
        cookies (list): list of cookies to pass to requests.post
    '''
    def __init__(self, node, url='', json=False, wrap=False, field=None, proxies=None, cookies=None):
        def _send(data, url=url, json=json, wrap=wrap, field=field, proxies=proxies, cookies=cookies):
            if wrap:
                data = [data]
            if json:
                data = JSON.dumps(data)

            msg = requests.post(url, data=data, cookies=cookies, proxies=proxies)

            if msg is None:
                return StreamEnd()

            if msg.status_code != 200:
                return msg

            if json:
                msg = msg.json()

            if field:
                msg = msg[field]

            if wrap:
                msg = [msg]

            return msg

        super().__init__(foo=_send, name='Http', inputs=1)
        node._downstream.append((self, 0))
        self._upstream.append(node)



import tributary.streaming as ts

def foo():
    yield 1
    yield 1
    yield 1
    yield 1
    yield 1


class TestRolling:
    def test_count(self):
        assert ts.run(ts.Count(ts.Foo(foo))) == [1, 2, 3, 4, 5]

    def test_sum(self):
        assert ts.run(ts.Sum(ts.Foo(foo))) == [1, 2, 3, 4, 5]



import math
from ..base import Node


def unary(foo, name):
    def _foo(self):
        downstream = Node(foo, {}, name=name, inputs=1)
        self._downstream.append((downstream, 0))
        downstream._upstream.append(self)
        return downstream
    return _foo


def binary(foo, name):
    def _foo(self, other):
        downstream = Node(foo, {}, name=name, inputs=2)
        self._downstream.append((downstream, 0))
        other._downstream.append((downstream, 1))
        downstream._upstream.extend([self, other])
        return downstream
    return _foo


########################
# Arithmetic Operators #
########################
Noop = unary(lambda x: x, name='Noop')
Negate = unary(lambda x: -1 * x, name='Negate')
Invert = unary(lambda x: 1 / x, name='Invert')
Add = binary(lambda x, y: x + y, name='Add')
Sub = binary(lambda x, y: x - y, name='Sub')
Mult = binary(lambda x, y: x * y, name='Mult')
Div = binary(lambda x, y: x / y, name='Div')
RDiv = binary(lambda x, y: y / x, name='RDiv')
Mod = binary(lambda x, y: x % y, name='Mod')
Pow = binary(lambda x, y: x ** y, name='Pow')


#####################
# Logical Operators #
#####################
Not = unary(lambda x: not x, name='Not')
And = binary(lambda x, y: x and y, name='And')
Or = binary(lambda x, y: x or y, name='Or')


###############
# Comparators #
###############
Equal = binary(lambda x, y: x == y, name='Equal')
NotEqual = binary(lambda x, y: x != y, name='NotEqual')
Lt = binary(lambda x, y: x < y, name='Less')
Le = binary(lambda x, y: x <= y, name='LessOrEqual')
Gt = binary(lambda x, y: x > y, name='Greater')
Ge = binary(lambda x, y: x >= y, name='GreaterOrEqual')


##########################
# Mathematical Functions #
##########################
Log = unary(lambda x: math.log(x), name='Log')
Sin = unary(lambda x: math.sin(x), name='Sin')
Cos = unary(lambda x: math.cos(x), name='Cos')
Tan = unary(lambda x: math.tan(x), name='Tan')
Arcsin = unary(lambda x: math.asin(x), name='Arcsin')
Arccos = unary(lambda x: math.acos(x), name='Arccos')
Arctan = unary(lambda x: math.atan(x), name='Arctan')
Sqrt = unary(lambda x: math.sqrt(x), name='Sqrt')
Abs = unary(lambda x: abs(x), name='Abs')
Exp = unary(lambda x: math.exp(x), name='Exp')
Erf = unary(lambda x: math.erf(x), name='Erf')


##############
# Converters #
##############
Int = unary(lambda x: int(x), name='Int')
Float = unary(lambda x: float(x), name='Float')
Bool = unary(lambda x: bool(x), name='Bool')
Str = unary(lambda x: str(x), name='Str')


###################
# Python Builtins #
###################
Len = unary(lambda x: len(x), name='Len')


########################
# Arithmetic Operators #
########################
Node.__add__ = Add
Node.__radd__ = Add
Node.__sub__ = Sub
Node.__rsub__ = Sub
Node.__mul__ = Mult
Node.__rmul__ = Mult
Node.__div__ = Div
Node.__rdiv__ = RDiv
Node.__truediv__ = Div
Node.__rtruediv__ = RDiv

Node.__pow__ = Pow
Node.__rpow__ = Pow
Node.__mod__ = Mod
Node.__rmod__ = Mod

#####################
# Logical Operators #
#####################
Node.__and__ = And
Node.__or__ = Or
Node.__invert__ = Not
Node.__bool__ = Bool
# TODO use __bool__ operator

##############
# Converters #
##############
Node.int = Int
Node.float = Float
Node.__str__ = Str

###############
# Comparators #
###############
Node.__lt__ = Lt
Node.__le__ = Le
Node.__gt__ = Gt
Node.__ge__ = Ge
Node.__eq__ = Equal
Node.__ne__ = NotEqual
Node.__neg__ = Negate
# Node.__nonzero__ = Bool  # Py2 compat

###################
# Python Builtins #
###################
Node.__len__ = Len

###################
# Numpy Functions #
###################
# Node.__array_ufunc__ = __array_ufunc__


##########################
# Mathematical Functions #
##########################
Node.log = Log
Node.sin = Sin
Node.cos = Cos
Node.tan = Tan
Node.asin = Arcsin
Node.acos = Arccos
Node.atan = Arctan
Node.abs = Abs
Node.sqrt = Sqrt
Node.exp = Exp
Node.erf = Erf



import aiofiles
import json as JSON
from ..base import Node


class File(Node):
    '''Open up a file and write lines to the file

    Args:
        node (Node): input stream
        filename (str): filename to write
        json (bool): load file line as json
    '''
    def __init__(self, node, filename='', json=True):
        async def _file(data):
            async with aiofiles.open(filename, mode='a') as fp:
                if json:
                    fp.write(JSON.dumps(data))
                else:
                    fp.write(data)
            return data

        super().__init__(foo=_file, name='File', inputs=1)
        node._downstream.append((self, 0))
        self._upstream.append(node)



from IPython.display import display
from ..base import Node


class Print(Node):
    def __init__(self, node, text=''):
        def foo(val):
            return val
        super().__init__(foo=foo, foo_kwargs=None, name='Print', inputs=1)

        node._downstream.append((self, 0))
        self._upstream.append(node)


def Graph(node):
    if not node._upstream:
        # leaf node
        return {node: []}
    return {node: [_.graph() for _ in node._upstream]}


def PPrint(node, level=0):
    ret = '    ' * (level - 1) if level else ''

    if not node._upstream:
        # leaf node
        return ret + '  \\  ' + repr(node)
    return '    ' * level + repr(node) + '\n' + '\n'.join(_.pprint(level + 1) for _ in node._upstream)


def GraphViz(node):
    d = Graph(node)

    from graphviz import Digraph
    dot = Digraph("Graph", strict=False)
    dot.format = 'png'

    def rec(nodes, parent):
        for d in nodes:
            if not isinstance(d, dict):
                dot.node(d)
                dot.edge(d, parent)

            else:
                for k in d:
                    dot.node(k._name)
                    rec(d[k], k)
                    dot.edge(k._name, parent._name)

    for k in d:
        dot.node(k._name)
        rec(d[k], k)

    return dot


class Perspective(Node):
    def __init__(self, node, text='', psp_kwargs=None):
        psp_kwargs = psp_kwargs or {}
        from perspective import PerspectiveWidget
        p = PerspectiveWidget(psp_kwargs.pop('schema', []), **psp_kwargs)

        def foo(val):
            p.update(val)
            return val
        super().__init__(foo=foo, foo_kwargs=None, name='Print', inputs=1)

        display(p)
        node._downstream.append((self, 0))
        self._upstream.append(node)
        self._name = "Perspective"


Node.graph = Graph
Node.pprint = PPrint
Node.graphviz = GraphViz
Node.print = Print
Node.perspective = Perspective



import tributary.lazy as t


class Foo1(t.LazyGraph):
    def __init__(self, *args, **kwargs):
        self.x = self.node('x', readonly=False, value=1, trace=True)


class Foo2(t.LazyGraph):
    def __init__(self, *args, **kwargs):
        self.y = self.node('y', readonly=False, value=2, trace=True)

        # ensure no __nodes clobber
        self.test = self.node('test', readonly=False, value=2, trace=True)
        self.x = self.node('x', readonly=False, value=2, trace=True)


class Foo3(t.LazyGraph):
    @t.node()
    def z(self):
        return self.x | self.y()

    @t.node()
    def y(self):
        return 10

    def reset(self):
        self.x = None

    def __init__(self):
        self.x = self.node(name="x", value=None)


class TestLazyOps:
    def test_add(self):
        f1 = Foo1()
        f2 = Foo2()
        z = f1.x + f2.y
        assert z() == 3
        f1.x = 2
        assert z() == 4
        f2.y = 4
        assert z() == 6

    def test_multi(self):
        f1 = Foo1()
        f2 = Foo1()
        f3 = Foo1()
        z = f1.x + f2.x - f3.x
        assert z() == 1
        f1.x = 2
        assert z() == 2
        f2.x = 4
        f3.x = 4
        assert z() == 2

    def test_or(self):
        f = Foo3()
        assert f.z()() == 10
        assert f.x() is None

        f.x = 5

        assert f.x() == 5
        assert f.z()() == 5

        f.reset()

        assert f.x() is None
        assert f.z()() == 10



from ..base import Node


class Count(Node):
    '''Node to count inputs'''
    def __init__(self, node, text=''):
        self._count = 0

        def foo(val):
            self._count += 1
            return self._count

        super().__init__(foo=foo, foo_kwargs=None, name='Count', inputs=1)

        node._downstream.append((self, 0))
        self._upstream.append(node)


class Sum(Node):
    '''Node to sum inputs'''
    def __init__(self, node, text=''):
        self._sum = 0

        def foo(val):
            self._sum += val
            return self._sum

        super().__init__(foo=foo, foo_kwargs=None, name='Sum', inputs=1)

        node._downstream.append((self, 0))
        self._upstream.append(node)



import tributary.symbolic as ts
import sympy as sy
from sympy.stats import Normal as syNormal, cdf


class TestConfig:
    def setup(self):
        pass
        # setup() before each test method

    def test_construct_lazy(self):
        # adapted from https://gist.github.com/raddy/bd0e977dc8437a4f8276
        spot, strike, vol, dte, rate, cp = sy.symbols('spot strike vol dte rate cp')

        T = dte / 260.
        N = syNormal('N', 0.0, 1.0)

        d1 = (sy.ln(spot / strike) + (0.5 * vol ** 2) * T) / (vol * sy.sqrt(T))
        d2 = d1 - vol * sy.sqrt(T)

        TimeValueExpr = sy.exp(-rate * T) * (cp * spot * cdf(N)(cp * d1) - cp * strike * cdf(N)(cp * d2))

        PriceClass = ts.construct_lazy(TimeValueExpr)

        price = PriceClass(spot=210.59, strike=205, vol=14.04, dte=4, rate=.2175, cp=-1)

        x = price.evaluate()()

        price.strike = 210

        assert x != price.evaluate()()

    def test_others(self):
        # adapted from https://gist.github.com/raddy/bd0e977dc8437a4f8276
        spot, strike, vol, dte, rate, cp = sy.symbols('spot strike vol dte rate cp')
        T = dte / 260.
        N = syNormal('N', 0.0, 1.0)
        d1 = (sy.ln(spot / strike) + (0.5 * vol ** 2) * T) / (vol * sy.sqrt(T))
        d2 = d1 - vol * sy.sqrt(T)
        TimeValueExpr = sy.exp(-rate * T) * (cp * spot * cdf(N)(cp * d1) - cp * strike * cdf(N)(cp * d2))
        PriceClass = ts.construct_lazy(TimeValueExpr)
        price = PriceClass(spot=210.59, strike=205, vol=14.04, dte=4, rate=.2175, cp=-1)
        price.evaluate()()
        ts.graphviz(TimeValueExpr)
        assert ts.traversal(TimeValueExpr)
        assert ts.symbols(TimeValueExpr)

    def test_parse(self):
        from sympy.parsing.sympy_parser import parse_expr
        assert parse_expr('x^2') == ts.parse_expression('x^2')

    # def test_construct_streaming(self):
    #     import tributary.symbolic as ts
    #     import tributary.reactive as tr

    #     expr = ts.parse_expression("10sin**2 x**2 + 3xyz + tan theta")
    #     clz = ts.construct_streaming(expr)

    #     def foo(*args):
    #         for i in range(10):
    #             yield i

    #     x = clz(x=tr.Const(1), y=tr.Foo(foo), z=tr.Timer(1, repeat=1), theta=tr.Const(4))
    #     x.pprint()
    #     out = x.run()
    #     assert len(out) == 1



from ..node import Node


def _print(node, counter=0, cache=None):
    if cache is None:
        cache = {}

    key = cache.get(id(node), str(node) + ' (#' + str(counter) + ')')
    cache[id(node)] = key

    if node._dirty or node._subtree_dirty() or node._always_dirty:
        key += '(dirty)'

    ret = {key: []}
    counter += 1

    if node._dependencies:
        for call, deps in node._dependencies.items():
            # callable node
            if hasattr(call, '_node_wrapper') and \
                    call._node_wrapper is not None:
                val, counter = call._node_wrapper._print(counter, cache)
                ret[key].append(val)

            # args
            for arg in deps[0]:
                val, counter = arg._print(counter, cache)
                ret[key].append(val)

            # kwargs
            for kwarg in deps[1].values():
                val, counter = kwarg._print(counter, cache)
                ret[key].append(val)

    return ret, counter


def Print(node):
    return node._print(0, {})[0]


def Graph(node):
    return node.print()


def GraphViz(node):
    d = node.graph()

    from graphviz import Digraph
    dot = Digraph(node._name, strict=True)
    dot.format = 'png'

    def rec(nodes, parent):
        for d in nodes:
            if not isinstance(d, dict):
                if '(dirty)' in d:
                    dot.node(d.replace('(dirty)', ''), color='red')
                    dot.edge(d.replace('(dirty)', ''), parent.replace('(dirty)', ''), color='red')
                else:
                    dot.node(d)
                    dot.edge(d, parent.replace('(dirty)', ''))
            else:
                for k in d:
                    if '(dirty)' in k:
                        dot.node(k.replace('(dirty)', ''), color='red')
                        rec(d[k], k)
                        dot.edge(k.replace('(dirty)', ''), parent.replace('(dirty)', ''), color='red')
                    else:
                        dot.node(k)
                        rec(d[k], k)
                        dot.edge(k, parent.replace('(dirty)', ''))

    for k in d:
        if '(dirty)' in k:
            dot.node(k.replace('(dirty)', ''), color='red')
        else:
            dot.node(k)
        rec(d[k], k)
    return dot


Node._print = _print
Node.print = Print
Node.graph = Graph
Node.graphviz = GraphViz



import asyncio
from .base import Node
from ..base import StreamNone, StreamRepeat


class Delay(Node):
    '''Streaming wrapper to delay a stream

    Arguments:
        node (node): input stream
        delay (float): time to delay input stream
    '''

    def __init__(self, node, delay=1):
        async def foo(val):
            await asyncio.sleep(delay)
            return val

        super().__init__(foo=foo, name='Delay', inputs=1)
        node._downstream.append((self, 0))
        self._upstream.append(node)


# class State(Node):
#     '''Streaming wrapper to delay a stream

#     Arguments:
#         node (node): input stream
#         state (dict): state dictionary of values to hold
#     '''

#     def __init__(self, node, delay=1):
#         async def foo(val):
#             await asyncio.sleep(delay)
#             return val

#         super().__init__(foo=foo, foo_kwargs=None, name='Delay', inputs=1)
#         node._downstream.append((self, 0))
#         self._upstream.append(node)


class Apply(Node):
    '''Streaming wrapper to apply a function to an input stream

    Arguments:
        node (node): input stream
        foo (callable): function to apply
        foo_kwargs (dict): kwargs for function
    '''
    def __init__(self, node, foo, foo_kwargs=None):
        self._apply = foo
        self._apply_kwargs = foo_kwargs or {}

        async def foo(val):
            return self._apply(val, **self._apply_kwargs)

        super().__init__(foo=foo, name='Apply', inputs=1)
        node._downstream.append((self, 0))
        self._upstream.append(node)


class Window(Node):
    '''Streaming wrapper to collect a window of values

    Arguments:
        node (node): input stream
        size (int): size of windows to use
        full_only (bool): only return if list is full
    '''
    def __init__(self, node, size=-1, full_only=False):
        self._accum = []

        def foo(val, size=size, full_only=full_only):
            if size == 0:
                return val
            else:
                self._accum.append(val)

            if size > 0:
                self._accum = self._accum[-size:]

            if full_only and len(self._accum) == size:
                return self._accum
            elif full_only:
                return StreamNone()
            else:
                return self._accum

        super().__init__(foo=foo, name='Window', inputs=1)
        node._downstream.append((self, 0))
        self._upstream.append(node)


class Unroll(Node):
    '''Streaming wrapper to unroll an iterable stream

    Arguments:
        node (node): input stream
    '''
    def __init__(self, node):
        self._count = 0

        async def foo(value):
            # unrolled
            if self._count > 0:
                self._count -= 1
                return value

            # unrolling
            try:
                for v in value:
                    self._count += 1
                    await self._push(v, 0)
            except TypeError:
                return value
            else:
                return StreamRepeat()

        super().__init__(foo=foo, name='Unroll', inputs=1)
        node._downstream.append((self, 0))
        self._upstream.append(node)


class UnrollDataFrame(Node):
    '''Streaming wrapper to unroll a dataframe into a stream

    Arguments:
        node (node): input stream
    '''
    def __init__(self, node, json=False, wrap=False):
        self._count = 0

        async def foo(value, json=json, wrap=wrap):
            # unrolled
            if self._count > 0:
                self._count -= 1
                return value

            # unrolling
            try:
                for i in range(len(value)):
                    row = value.iloc[i]

                    if json:
                        data = row.to_dict()
                        data['index'] = row.name
                    else:
                        data = row
                    self._count += 1
                    await self._push(data, 0)

            except TypeError:
                return value
            else:
                return StreamRepeat()

        super().__init__(foo=foo, name='UnrollDF', inputs=1)
        node._downstream.append((self, 0))
        self._upstream.append(node)


class Merge(Node):
    '''Streaming wrapper to merge 2 inputs into a single output

    Arguments:
        node1 (node): input stream
        node2 (node): input stream
    '''
    def __init__(self, node1, node2):
        def foo(value1, value2):
            return value1, value2

        super().__init__(foo=foo, name='Merge', inputs=2)
        node1._downstream.append((self, 0))
        node2._downstream.append((self, 1))
        self._upstream.append(node1)
        self._upstream.append(node2)


class ListMerge(Node):
    '''Streaming wrapper to merge 2 input lists into a single output list

    Arguments:
        node1 (node): input stream
        node2 (node): input stream
    '''
    def __init__(self, node1, node2):
        def foo(value1, value2):
            return list(value1) + list(value2)

        super().__init__(foo=foo, name='ListMerge', inputs=2)
        node1._downstream.append((self, 0))
        node2._downstream.append((self, 1))
        self._upstream.append(node1)
        self._upstream.append(node2)


class DictMerge(Node):
    '''Streaming wrapper to merge 2 input dicts into a single output dict.
       Preference is given to the second input (e.g. if keys overlap)

    Arguments:
        node1 (node): input stream
        node2 (node): input stream
    '''
    def __init__(self, node1, node2):
        def foo(value1, value2):
            ret = {}
            ret.update(value1)
            ret.update(value2)
            return ret

        super().__init__(foo=foo, name='DictMerge', inputs=2)
        node1._downstream.append((self, 0))
        node2._downstream.append((self, 1))
        self._upstream.append(node1)
        self._upstream.append(node2)


class Reduce(Node):
    '''Streaming wrapper to merge any number of inputs

    Arguments:
        nodes (tuple): input streams
    '''
    def __init__(self, *nodes):
        def foo(*values):
            return values

        super().__init__(foo=foo, name='Reduce', inputs=len(nodes))
        for i, n in enumerate(nodes):
            n._downstream.append((self, i))
            self._upstream.append(n)



from gevent import monkey
_PATCHED = False

if not _PATCHED:
    monkey.patch_all(thread=False, select=False)
    _PATCHED = True


from ._version import __version__  # noqa: F401
from .functional import pipeline, stop  # noqa: F401



import tornado.websocket
import tornado.web
import tornado.ioloop
import time
from tributary.streaming.input import _gen


class DummyWebSocket(tornado.websocket.WebSocketHandler):
    def open(self):
        print("WebSocket opened")
        i = 0
        x = {y: _gen() for y in ('A', 'B', 'C', 'D')}
        try:
            while i < len(x['A']):
                self.write_message({'A': x['A'][i],
                                    'B': x['B'][i],
                                    'C': x['C'][i],
                                    'D': x['D'][i]})
                i += 1
                time.sleep(.1)
        finally:
            print("WebSocket closed")
            self.close()

    def on_message(self, message):
        self.write_message(u"You said: " + message)

    def on_close(self):
        print("WebSocket closed")


def main():
    app = tornado.web.Application([(r"/", DummyWebSocket)])
    app.listen(8899)
    print('listening on %d' % 8899)
    tornado.ioloop.IOLoop.current().start()


if __name__ == '__main__':
    main()



import math
import numpy as np
import scipy as sp
from ..base import Node


########################
# Arithmetic Operators #
########################
def Add(self, other):
    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '+' + other._name, (lambda x, y: x.value() + y.value()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '+' + other._name, (lambda x, y: x.value() + y.value()), [self, other], self._trace or other._trace)


def Sub(self, other):
    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '-' + other._name, (lambda x, y: x.value() - y.value()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '-' + other._name, (lambda x, y: x.value() - y.value()), [self, other], self._trace or other._trace)


def Mult(self, other):
    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '*' + other._name, (lambda x, y: x.value() * y.value()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '*' + other._name, (lambda x, y: x.value() * y.value()), [self, other], self._trace or other._trace)


def Div(self, other):
    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '/' + other._name, (lambda x, y: x.value() / y.value()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '/' + other._name, (lambda x, y: x.value() / y.value()), [self, other], self._trace or other._trace)


def Pow(self, other):
    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '^' + other._name, (lambda x, y: x.value() ** y.value()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '^' + other._name, (lambda x, y: x.value() ** y.value()), [self, other], self._trace or other._trace)


def Negate(self):
    return self._gennode('(-' + self._name + ')', (lambda x: -self.value()), [self], self._trace)


#####################
# Logical Operators #
#####################
def Or(self, other):
    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '||' + other._name, (lambda x, y: x.value() or y.value()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '||' + other._name, (lambda x, y: x.value() or y.value()), [self, other], self._trace or other._trace)


def And(self, other):
    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '&&' + other._name, (lambda x, y: x.value() or y.value()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '&&' + other._name, (lambda x, y: x.value() and y.value()), [self, other], self._trace or other._trace)


def Not(self):
    return self._gennode('!' + self._name, (lambda x: not x.value()), [self], self._trace)


##########################
# Mathematical Functions #
##########################
def Sin(self):
    return self._gennode('sin(' + self._name + ')', (lambda x: math.sin(self.value())), [self], self._trace)


def Cos(self):
    return self._gennode('cos(' + self._name + ')', (lambda x: math.cos(self.value())), [self], self._trace)


def Tan(self):
    return self._gennode('tan(' + self._name + ')', (lambda x: math.tan(self.value())), [self], self._trace)


def Arcsin(self):
    return self._gennode('sin(' + self._name + ')', (lambda x: math.arcsin(self.value())), [self], self._trace)


def Arccos(self):
    return self._gennode('arccos(' + self._name + ')', (lambda x: math.arccos(self.value())), [self], self._trace)


def Arctan(self):
    return self._gennode('arctan(' + self._name + ')', (lambda x: math.arctan(self.value())), [self], self._trace)


def Abs(self):
    return self._gennode('||' + self._name + '||', (lambda x: abs(self.value())), [self], self._trace)


def Sqrt(self):
    return self._gennode('sqrt(' + self._name + ')', (lambda x: math.sqrt(self.value())), [self], self._trace)


def Log(self):
    return self._gennode('log(' + self._name + ')', (lambda x: math.log(self.value())), [self], self._trace)


def Exp(self):
    return self._gennode('exp(' + self._name + ')', (lambda x: math.exp(self.value())), [self], self._trace)


def Erf(self):
    return self._gennode('erf(' + self._name + ')', (lambda x: math.erf(self.value())), [self], self._trace)


##############
# Converters #
##############
def Float(self):
    return self._gennode('float(' + self._name + ')', (lambda x: float(self.value())), [self], self._trace)


def Int(self):
    return self._gennode('int(' + self._name + ')', (lambda x: int(self.value())), [self], self._trace)


def Bool(self):
    if self.value() is None:
        return False
    return bool(self.value())


###################
# Python Builtins #
###################
def Len(self):
    return self._gennode('len(' + self._name + ')', (lambda x: len(self.value())), [self], self._trace)


###################
# Numpy Functions #
###################
def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
    if ufunc == np.add:
        if isinstance(inputs[0], Node):
            return inputs[0].__add__(inputs[1])
        else:
            return inputs[1].__add__(inputs[0])
    elif ufunc == np.subtract:
        if isinstance(inputs[0], Node):
            return inputs[0].__sub__(inputs[1])
        else:
            return inputs[1].__sub__(inputs[0])
    elif ufunc == np.multiply:
        if isinstance(inputs[0], Node):
            return inputs[0].__mul__(inputs[1])
        else:
            return inputs[1].__mul__(inputs[0])
    elif ufunc == np.divide:
        if isinstance(inputs[0], Node):
            return inputs[0].__truedivide__(inputs[1])
        else:
            return inputs[1].__truedivide__(inputs[0])
    elif ufunc == np.sin:
        return inputs[0].sin()
    elif ufunc == np.cos:
        return inputs[0].cos()
    elif ufunc == np.tan:
        return inputs[0].tan()
    elif ufunc == np.arcsin:
        return inputs[0].arcsin()
    elif ufunc == np.arccos:
        return inputs[0].arccos()
    elif ufunc == np.arctan:
        return inputs[0].arctan()
    elif ufunc == np.exp:
        return inputs[0].exp()
    elif ufunc == sp.special.erf:
        return inputs[0].erf()
    else:
        raise NotImplementedError('Not Implemented!')


###############
# Comparators #
###############
def Equal(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return True

    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '==' + other._name, (lambda x, y: x() == y()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '==' + other._name, (lambda x, y: x() == y()), [self, other], self._trace or other._trace)


def NotEqual(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return False

    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '!=' + other._name, (lambda x, y: x() != y()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '!=' + other._name, (lambda x, y: x() != y()), [self, other], self._trace or other._trace)


def Ge(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return True

    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '>=' + other._name, (lambda x, y: x() >= y()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '>=' + other._name, (lambda x, y: x() >= y()), [self, other], self._trace or other._trace)


def Gt(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return False

    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '>' + other._name, (lambda x, y: x() > y()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '>' + other._name, (lambda x, y: x() > y()), [self, other], self._trace or other._trace)


def Le(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return True

    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '<=' + other._name, (lambda x, y: x() <= y()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '<=' + other._name, (lambda x, y: x() <= y()), [self, other], self._trace or other._trace)


def Lt(self, other):
    if isinstance(other, Node) and super(Node, self).__eq__(other):
        return False
    other = self._tonode(other)
    if isinstance(self._self_reference, Node):
        return self._gennode(self._name + '<' + other._name, (lambda x, y: x() < y()), [self._self_reference, other], self._trace or other._trace)
    return self._gennode(self._name + '<' + other._name, (lambda x, y: x() < y()), [self, other], self._trace or other._trace)


########################
# Arithmetic Operators #
########################
Node.__add__ = Add
Node.__radd__ = Add
Node.__sub__ = Sub
Node.__rsub__ = Sub
Node.__mul__ = Mult
Node.__rmul__ = Mult
Node.__div__ = Div
Node.__rdiv__ = Div
Node.__truediv__ = Div
Node.__rtruediv__ = Div

Node.__pow__ = Pow
Node.__rpow__ = Pow
# Node.__mod__ = Mod
# Node.__rmod__ = Mod

#####################
# Logical Operators #
#####################
Node.__and__ = And
Node.__or__ = Or
Node.__invert__ = Not

##############
# Converters #
##############
Node.int = Int
Node.float = Float
Node.__bool__ = Bool

###############
# Comparators #
###############
Node.__lt__ = Lt
Node.__le__ = Le
Node.__gt__ = Gt
Node.__ge__ = Ge
Node.__eq__ = Equal
Node.__ne__ = NotEqual
Node.__neg__ = Negate
Node.__nonzero__ = Bool  # Py2 compat

###################
# Python Builtins #
###################
Node.__len__ = Len

###################
# Numpy Functions #
###################
Node.__array_ufunc__ = __array_ufunc__

##########################
# Mathematical Functions #
##########################
Node.log = Log
Node.sin = Sin
Node.cos = Cos
Node.tan = Tan
Node.arcsin = Arcsin
Node.arccos = Arccos
Node.arctan = Arctan
Node.abs = Abs
Node.sqrt = Sqrt
Node.exp = Exp
Node.erf = Erf



import json as JSON
import websockets
from ..base import Node
from ...base import StreamNone, StreamEnd


class WebSocket(Node):
    '''Connect to websocket and send data

    Args:
        foo (callable): input stream
        foo_kwargs (dict): kwargs for the input stream
        url (str): websocket url to connect to
        json (bool): dump data as json
        wrap (bool): wrap result in a list
    '''
    def __init__(self, node, url='', json=False, wrap=False, field=None, response=False):
        self._websocket = websockets.connect(url)

        async def _send(data, url=url, json=json, wrap=wrap, field=field, response=response):
            if isinstance(data, (StreamNone, StreamEnd)):
                return data

            if wrap:
                data = [data]
            if json:
                data = JSON.dumps(data)

            await self._websocket.send(data)

            if response:
                msg = await self._websocket.recv()

            else:
                msg = '{}'

            if json:
                msg = JSON.loads(msg)

            if field:
                msg = msg[field]

            if wrap:
                msg = [msg]

            return msg

        super().__init__(foo=_send, name='WebSocket', inputs=1)
        node._downstream.append((self, 0))
        self._upstream.append(node)
